﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Decal.Adapter.Wrappers;
using VirindiViewService;
using VirindiViewService.Controls;
using System.Windows.Forms;

namespace CompHud
{
    public partial class PluginCore
    {

        private HudView hud;
        private HudFixedLayout head;

        private const int LINEHEIGHT = 14;
        private const int TEXTSIZE = 9;
        private const int DEFAULTICON = 0x06002D64;

        private Dictionary<string, HudStaticText> countControls;
        private List<HudControl> hudControls;
        private Timer timer;

        private void createHud()
        {
            if (inactive || pluginSettings == null) return;

            if (hud == null)
            {
                hud = new HudView(PLUGIN, 200, 20, new ACImage(System.Drawing.Color.Black), false);
                hud.Alpha = pluginSettings.alpha;
                hud.Icon = 0x64F9;
                hud.Visible = true;
                hud.UserGhostable = true;
                hud.Ghosted = true;
                hud.UserMinimizable = false;
                hud.UserAlphaChangeable = false;
                hud.ShowIcon = true;
                hud.Location = new Point(0, 180);

                hud.LoadUserSettings();

                head = new HudFixedLayout();
                hud.Controls.HeadControl = head;

                countControls = new Dictionary<string, HudStaticText>();
                hudControls = new List<HudControl>();
            }
        }

        private void drawHud()
        {
            if (inactive || pluginSettings == null) return;

            if (hud == null || countControls == null || hudControls == null) createHud();

            // stop the timer
            if (timer != null && timer.Enabled) stopTimer();
            
            hud.Visible = pluginSettings.visible;

            if (hud.Visible)
            {
                hud.Alpha = pluginSettings.alpha;
                hud.Width = pluginSettings.textOff ? 60 : 200;
                hud.ThemeChanged += new EventHandler(hud_ThemeChanged);

                int h = 0;

                foreach (HudControl ctrl in hudControls)
                {
                    head.RemovedChild(ctrl);
                    ctrl.Dispose();
                }
                hudControls.Clear();
                countControls.Clear();

                foreach (KeyValuePair<string, int> kv in pluginSettings.WatchedItems)
                {
                    hud.Height = h + LINEHEIGHT + LINEHEIGHT - TEXTSIZE + 2;

                    HudPictureBox bx = new HudPictureBox();
                    head.AddControl(bx, new Rectangle(0, h, 16, 16));
                    hudControls.Add(bx);
                    if (kv.Value >= 0)
                    {
                        bx.Image = new ACImage(kv.Value > 0 ? kv.Value : DEFAULTICON);
                    }

                    int itemCount = getItemCount(kv.Key);

                    HudStaticText ax = new HudStaticText();
                    ax.FontHeight = TEXTSIZE;
                    ax.Text = formatNumber(itemCount);
                    ax.TextColor = getTextColor(kv.Key, itemCount);

                    head.AddControl(ax, new Rectangle(20, h, 40, 16));
                    countControls.Add(kv.Key, ax);
                    hudControls.Add(ax);

                    if (!pluginSettings.textOff)
                    {
                        HudStaticText tx = new HudStaticText();
                        tx.FontHeight = TEXTSIZE;
                        tx.Text = kv.Key;
                        tx.TextColor = ax.TextColor;
                        head.AddControl(tx, new Rectangle(60, h, 140, 16));
                        hudControls.Add(tx);
                    }

                    h += LINEHEIGHT + LINEHEIGHT - TEXTSIZE;
                }
                startTimer();
            }
        }

        void hud_ThemeChanged(object sender, EventArgs e)
        {
            refreshHud();
        }

        private void refreshHud()
        {
            if (inactive || hud == null) return;

            foreach (KeyValuePair<string, HudStaticText> kv in countControls)
            {
                int itemCount = getItemCount(kv.Key);
                kv.Value.Text = formatNumber(itemCount);
                kv.Value.TextColor = getTextColor(kv.Key, itemCount);
            }
        }

        private Color getTextColor(string item, int itemCount)
        {
            return (itemCount < 4 || (item.Contains("Taper") && itemCount < 100)) ? hud.Theme.GetColor("Hint_ListTextSelectionColor") : hud.Theme.GetColor("ViewText");
        }

        private void startTimer()
        {
            if (timer == null)
            {
                timer = new Timer();
                timer.Interval = 5000;
                timer.Tick += new EventHandler(timer_Tick);
            }
            if (!timer.Enabled) timer.Start();
        }

        void timer_Tick(object sender, EventArgs e)
        {
            if (inactive || hud == null) return;

            try
            {
                refreshHud();

                foreach (DelayedChatParserText dcpt in timedChatParserCommands) {
                    if (dcpt.isDue(8))
                    {
                        Host.Actions.InvokeChatParser(dcpt.command);
                        dcpt.command = String.Empty;
                    }
                }
                timedChatParserCommands.RemoveAll(delegate(DelayedChatParserText d) { return String.Empty.Equals(d.command); });
            }
            catch (Exception ex) { }
        }

        private void stopTimer()
        {
            if (timer != null)
            {
                timer.Stop();
                timer.Dispose();
                timer = null;
            }
        }

        private string formatNumber(int n)
        {
            if (n > 9999)
            {
                return string.Format("{0}k", Math.Floor((double)n/1000));
            }
            return n.ToString();
        }

    }
}
