﻿using System;
using System.IO;
using System.Windows.Forms;
using Decal.Adapter;
using Decal.Adapter.Wrappers;
using System.Collections.Generic;
using System.Text.RegularExpressions;

namespace CompHud
{
    delegate void moveTimer_Tick(object sender, EventArgs e);

    public delegate string getSortKey(WorldObject wo);

    public partial class PluginCore
    {
        public class ItemMoveInstruction
        {
            public int guid;
            public int pack;
            public int slot;

            public static ItemMoveInstruction get(int guid, int pack, int slot)
            {
                ItemMoveInstruction i = new ItemMoveInstruction();
                i.guid = guid;
                i.pack = pack;
                i.slot = slot;
                return i;
            }
        }

        private List<ItemMoveInstruction> itemsToMove = null;
        private Timer moveTimer;

        private bool canMoveItems()
        {
            if (inactive || pluginSettings == null) return false;

            if (itemsToMove != null)
            {
                WriteToChat("Still busy. Type '/ch stop' to abort.");
                return false;
            }

            return true;
        }

        public int getPackId(int c)
        {
            c = Math.Max(0, Math.Min(9, c));
            if (c == 0)
            {
                return Core.CharacterFilter.Id;
            }
            else
            {
                foreach (WorldObject wo in Core.WorldFilter.GetInventory())
                {
                    if (wo.ObjectClass == ObjectClass.Container)
                    {
                        //WriteToChat(String.Format("{0} is a container: {1} {2} {3}", wo.Name, wo.Id, wo.Container, wo.Values(LongValueKey.Slot, 0)));
                        if (wo.Values(LongValueKey.Slot, 0) == c - 1)
                        {
                            return wo.Id;
                        }
                    }
                }
            }
            return 0;
        }

        public void tryCramToContainer(String c, ItemMatcher matcher)
        {
            try
            {
                moveItems(0, Convert.ToInt32(c), matcher);
            }
            catch (Exception ex)
            {
                WriteToChat("Invalid command: " + ex.Message);
            }
        }

        public void tryUnCramFromContainer(String c, ItemMatcher matcher)
        {
            try
            {
                moveItems(Convert.ToInt32(c), 0, matcher);
            }
            catch (Exception ex)
            {
                WriteToChat("Invalid command: " + ex.Message);
            }
        }

        public void moveItems(int from, int to, ItemMatcher matcher)
        {
            if (!canMoveItems() || from == to) return;
            int fromId = getPackId(from);
            int toId = getPackId(to);

            itemsToMove = new List<ItemMoveInstruction>();

            foreach (WorldObject wo in Core.WorldFilter.GetInventory())
            {
                if (wo.ObjectClass != ObjectClass.Container && wo.ObjectClass != ObjectClass.Foci)
                {
                    if (matcher.IsMatch(wo) && isInPack(wo, fromId, true))
                    {
                        itemsToMove.Add(ItemMoveInstruction.get(wo.Id, toId, 0));
                    }
                }
            }
            string msg = String.Format("Moving items matching '{0}' to container {1}", matcher, to);
            mayStartMoveTimer(msg, false);
        }

        public void sortContainer(int c, getSortKey gsk)
        {
            if (!canMoveItems()) return;

            c = Math.Max(0, Math.Min(9, c));

            SortedDictionary<String, WorldObject> items = new SortedDictionary<String, WorldObject>();
            int pack = getPackId(c);
            int missingIdData = 0;
            if (pack != 0)
            {
                int cnt = 1000;
                try
                {
                    foreach (WorldObject wo in Core.WorldFilter.GetInventory())
                    {
                        if (missingIdData > 100)
                        {
                            break;
                        }

                        if (isInPack(wo, pack, true) && wo.ObjectClass != ObjectClass.Container && wo.ObjectClass != ObjectClass.Foci)
                        {
                            string sk = gsk(wo);
                            if (sk == null)
                            {
                                missingIdData += 1;
                            }
                            else
                            {
                                if (items.ContainsKey(sk))
                                {
                                    items.Add(String.Format("{0}-{1}", sk, cnt++), wo);
                                }
                                else
                                {
                                    items.Add(sk, wo);
                                }
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    WriteToChat("Unexpected exception: " + ex.Message);
                }
            }

            itemsToMove = new List<ItemMoveInstruction>();
            foreach (KeyValuePair<String, WorldObject> kv in items)
            {
                itemsToMove.Add(ItemMoveInstruction.get(kv.Value.Id, pack, itemsToMove.Count));
            }

            mayStartMoveTimer(String.Format("Sorting container {0}", c), missingIdData > 0);
        }

        private void mayStartMoveTimer(string msg, bool missingIdData)
        {
            if (missingIdData)
            {
                WriteToChat("ID data missing for some items. Please try again in a few seconds.");
                itemsToMove = null;
            }
            else
            {
                WriteToChat(String.Format("{0} ({1} items)", msg, itemsToMove.Count));
                startMoveTimer();
            }
        }

        string defaultSortKey(WorldObject wo)
        {
            if (wo.ObjectClass == ObjectClass.SpellComponent)
            {
                return spellcompSortKey(wo);
            }
            else if (isSalvage(wo))
            {
                return string.Format("__SALVAGE: {0} {1}", getMaterialName(wo.Values(LongValueKey.Material, 0)), 100 + wo.Values(DoubleValueKey.SalvageWorkmanship, 0));
            }
            else if (isEquipableItem(wo))
            {
                if (wo.HasIdData)
                {
                    if (isWeapon(wo))
                    {
                        return string.Format("_WEAPON: {0} {1} {2}", wo.Values(LongValueKey.EquipSkill, 0), 999 - wo.Values(LongValueKey.WieldReqValue, 0),
                            wo.Values(LongValueKey.DamageType, 0));
                    }
                    if (isWand(wo))
                    {
                        return string.Format("_WAND: {0} {1} {2}", wo.Values(LongValueKey.WieldReqAttribute, 0), 999 - wo.Values(LongValueKey.WieldReqValue, 0),
                            wo.Values(LongValueKey.WandElemDmgType, 0));
                    }
                    return string.Format("_{0}: {1} {2} {3}", getEffectiveObjectClass(wo), wo.Values(LongValueKey.EquipableSlots, 0), getSpellNameForSorting(wo),
                        1000 + wo.Values(LongValueKey.LoreRequirement, 0));
                }
                else
                {
                    return null;
                }
            }
            return objectClassSortKey(wo);
        }

        string tinkSortKey(WorldObject wo)
        {
            if (isArmor(wo) || isWandOrWeapon(wo))
            {
                if (wo.HasIdData)
                {
                    return string.Format("_TINKED {0}", 10 - wo.Values(LongValueKey.NumberTimesTinkered));
                }
                else
                {
                    return null;
                }
            }
            return objectClassSortKey(wo);
        }

        string valueSortKey(WorldObject wo)
        {
            return string.Format("{0} {1}", 999999999 - getSingleValue(wo), wo.Name);
        }

        string setSortKey(WorldObject wo)
        {
            if (wo.HasIdData)
            {
                int setId = WorldObjectInfo.getItemsSetId(wo);
                return string.Format("{0} {1}", WorldObjectInfo.getSetName(setId), defaultSortKey(wo));
            }
            else
            {
                return null;
            }
        }

        string spellcompSortKey(WorldObject wo)
        {
            if (wo.Name.EndsWith("Pea") || wo.Name.EndsWith("Pyramid"))
            {
                if (Array.IndexOf(SPELLCOMPS_ALCH_PEA, wo.Name) > -1)
                {
                    return string.Format("_COMP XA {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_HERB_PEA, wo.Name) > -1)
                {
                    return string.Format("_COMP XH {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_POWDER_PEA, wo.Name) > -1)
                {
                    return string.Format("_COMP XP {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_TALISMAN_PEA, wo.Name) > -1)
                {
                    return string.Format("_COMP XT {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_TAPER_PEA, wo.Name) > -1)
                {
                    return string.Format("_COMP XU {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_SCARAB_PEA, wo.Name) > -1)
                {
                    return string.Format("_COMP XZ {0}", valueSortKey(wo));
                }
            }            
            else
            {
                if (Array.IndexOf(SPELLCOMPS_ALCH, wo.Name) > -1)
                {
                    return string.Format("_COMP RA {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_HERB, wo.Name) > -1)
                {
                    return string.Format("_COMP RH {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_POWDER, wo.Name) > -1)
                {
                    return string.Format("_COMP RP {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_TALISMAN, wo.Name) > -1)
                {
                    return string.Format("_COMP RT {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_TAPER, wo.Name) > -1)
                {
                    return string.Format("_COMP RU {0}", wo.Name);
                }
                else if (Array.IndexOf(SPELLCOMPS_SCARAB, wo.Name) > -1)
                {
                    return string.Format("_COMP RZ {0}", valueSortKey(wo));
                }
            }
            return string.Format("_COMP ZZ {1}", valueSortKey(wo));
        }

        string objectClassSortKey(WorldObject wo)
        {
            return string.Format("{0} {1}", getEffectiveObjectClass(wo), wo.Name);
        }

        private void startMoveTimer()
        {
            stopMoveTimer();
            moveTimer = new Timer();
            moveTimer.Interval = 300;
            moveTimer.Tick += new EventHandler(moveTimer_Tick);
            moveTimer.Start();
        }

        private void stopMoveTimer()
        {
            if (moveTimer != null)
            {
                moveTimer.Stop();
                moveTimer.Dispose();
                moveTimer = null;
            }
        }

        void moveTimer_Tick(object sender, EventArgs e)
        {
            if (Host.Actions.BusyState != 0) { return; }

            try
            {
                ItemMoveInstruction imi = itemsToMove[0];

                if (imi == null)
                {
                    throw new Exception();
                }
                else
                {
                    //WriteToChat(String.Format("Moving item {0:x4} from {2} to {1}", imi.guid, imi.slot, Core.WorldFilter[imi.guid].Values(LongValueKey.Slot, 0)));
                    Host.Actions.MoveItem(imi.guid, imi.pack, imi.slot < 0 ? 0 : imi.slot, false);
                    itemsToMove.Remove(imi);
                }
            }
            catch (Exception ex)
            {
                stopAll();
            }
        }

        private void stopAll()
        {
            stopMoveTimer();
            itemsToMove = null;
            WriteToChat("All sorting stopped.");
        }
    }
}